Package: codez
Type: Package
Title: Seq2Seq Encoder-Decoder Model for Time-Feature Analysis Based on
        Tensorflow
Version: 1.0.0
Authors@R: c(person("Giancarlo", "Vercellino", role = c("aut", "cre", "cph"), email = "giancarlo.vercellino@gmail.com"))
Maintainer: Giancarlo Vercellino <giancarlo.vercellino@gmail.com>
Description: Proposes Seq2seq Time-Feature Analysis using an Encoder-Decoder to project into latent space and a Forward Network to predict the next sequence.
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.2.1
Depends: R (>= 3.6)
Imports: purrr (>= 0.3.4), abind (>= 1.4-5), ggplot2 (>= 3.3.6), readr
        (>= 2.1.2), fANCOVA (>= 0.6-1), imputeTS (>= 3.2), modeest (>=
        2.4.0), scales (>= 1.1.1), tictoc (>= 1.0.1), tensorflow (>=
        2.9.0), keras (>= 2.9.0), moments (>= 0.14), narray (>=
        0.4.1.1), fastDummies (>= 1.6.3), entropy (>= 1.3.1),
        philentropy (>= 0.5.0), greybox (>= 1.0.1), lubridate (>=
        1.7.10)
URL: https://rpubs.com/giancarlo_vercellino/codez
Suggests: testthat (>= 3.0.0), reticulate (>= 1.26)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2022-09-23 08:43:28 UTC; gvercellino
Author: Giancarlo Vercellino [aut, cre, cph]
Repository: CRAN
Date/Publication: 2022-09-23 15:10:02 UTC
Built: R 4.2.0; ; 2023-07-11 02:31:52 UTC; unix
