Package: codez
Type: Package
Title: Torch Seq2Seq Encoder-Decoder Model for Time-Feature Analysis
Version: 2.0.0
Authors@R: c(person("Giancarlo", "Vercellino", role = c("aut", "cre", "cph"), email = "giancarlo.vercellino@gmail.com"))
Maintainer: Giancarlo Vercellino <giancarlo.vercellino@gmail.com>
Description: Proposes Seq2seq Time-Feature Analysis using a torch Encoder-Decoder to project into latent space and a Forward Network to predict the next sequence, with dependency-light local support functions, tidy outputs and baseline backtesting helpers.
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Depends: R (>= 3.6)
URL: https://rpubs.com/giancarlo_vercellino/codez
Suggests: testthat (>= 3.0.0), torch
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-07-17 14:12:18 UTC; gianc
Author: Giancarlo Vercellino [aut, cre, cph]
Repository: CRAN
Date/Publication: 2026-07-17 15:50:21 UTC
