#' amzn_aapl_fb data set
#'
#' A data frame with the close prices for Amazon, Google and Facebook.
#'
#' @format A data frame with 4 columns and 1798 rows.
#'
#' @source Yahoo Finance
#'
"amzn_aapl_fb"
