#' codez
#'
#' Seq2seq Time-Feature Analysis using an Encoder-Decoder to project into latent space and a Forward Network to predict the next sequence.
#'
#' @docType package
#' @name codez
"_PACKAGE"
