\name{autoplot}
\alias{autoplot}
\title{Create a codez plot object}
\usage{
autoplot(object, ...)
}
\arguments{
\item{object}{An object with an \code{autoplot()} method.}
\item{...}{Additional arguments passed to methods.}
}
\value{
A lightweight codez plot object when called on a supported codez object.
}
\description{
Generic for creating dependency-free plot objects. Use \code{plot()} to draw
the returned object with base R graphics.
}
